# -*-cmake-*-
#
# Try to find the libzoltan graph partioning library
#
# Once done, this will define:
#
#  ZOLTAN_FOUND        - system has the libzoltan graph partioning library
#  HAVE_ZOLTAN         - like ZOLTAN_FOUND, but for the inclusion in config.h
#  ZOLTAN_INCLUDE_DIR  - incude paths to use libzoltan
#  ZOLTAN_LIBRARIES    - Link these to use libzoltan
set(ZOLTAN_SEARCH_PATH "/usr" "/usr/local" "/opt" "/opt/local")
set(ZOLTAN_NO_DEFAULT_PATH "")
if(ZOLTAN_ROOT)
  set(ZOLTAN_SEARCH_PATH "${ZOLTAN_ROOT}")
  set(ZOLTAN_NO_DEFAULT_PATH "NO_DEFAULT_PATH")
endif()

# Make sure we have checked for the underlying partitioners.
find_package(PTScotch)

# search for files which implements this module
find_path (ZOLTAN_INCLUDE_DIRS
  NAMES "zoltan.h"
  PATHS ${ZOLTAN_SEARCH_PATH}
  PATH_SUFFIXES "include"
  ${ZOLTAN_NO_DEFAULT_PATH})

message("Found ${ZOLTAN_INCLUDE_DIRS}")

# only search in architecture-relevant directory
if (CMAKE_SIZEOF_VOID_P)
  math (EXPR _BITS "8 * ${CMAKE_SIZEOF_VOID_P}")
endif (CMAKE_SIZEOF_VOID_P)

find_library(ZOLTAN_LIBRARIES
  NAMES "zoltan"
  PATHS ${ZOLTAN_SEARCH_PATH}
  PATH_SUFFIXES "lib/.libs" "lib" "lib${_BITS}" "lib/${CMAKE_LIBRARY_ARCHITECTURE}"
  ${ZOLTAN_NO_DEFAULT_PATH})

message("Found ${ZOLTAN_LIBRARIES}")

set (ZOLTAN_FOUND FALSE)
if (ZOLTAN_INCLUDE_DIRS OR ZOLTAN_LIBRARIES)
  set(ZOLTAN_FOUND TRUE)
  set(HAVE_ZOLTAN 1)
  set(ZOLTAN_LIBRARIES ${ZOLTAN_LIBRARIES} ${PARMETIS_LIBRARIES} ${PTSCOTCH_LIBRARIES})
  set(ZOLTAN_INCLUDE_DIRS ${ZOLTAN_INCLUDE_DIRS} ${PARMETIS_INCLUDE_DIRS}
      ${PTSCOTCH_INCLUDE_DIRS})

  # register all ZOLTAN related flags
  dune_register_package_flags(LIBRARIES "${ZOLTAN_LIBRARIES}"
                            INCLUDE_DIRS "${ZOLTAN_INCLUDE_DIRS}")
  set (ZOLTAN_CONFIG_VAR HAVE_ZOLTAN)
endif()

# print a message to indicate status of this package
#include (FindPackageHandleStandardArgs)
#find_package_handle_standard_args(ZOLTAN
#  DEFAULT_MSG
#  ZOLTAN_LIBRARIES
#  ZOLTAN_INCLUDE_DIRS
#  )
